unit NeuralNet2;

// Neural MLP network
//
// (c)2003, TEDSoft GmbH, Bonn, Germany
//
// www.tedsoft.de, info@tedsoft.de
//
// Alle Rechte vorbehalten.
// 
// All rights reserved - unless granted by written agreement.
//
// It is not permitted to remove this notice.
//
// Free for noncommercial use.

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, TedMLPNet, StdCtrls, TedNetUnit;

type
  TmlpDemoForm = class(TForm)
    InitialisingGroupBox: TGroupBox;
    OpenDialog: TOpenDialog;
    trainingButton: TButton;
    LearningGroupBox: TGroupBox;
    LearningButton: TButton;
    numbOfCyclesEdit: TEdit;
    numbOfCyclesLabel: TLabel;
    errorCheckBox: TCheckBox;
    ErrorBoundEdit: TEdit;
    ErrorBoundLabel: TLabel;
    VerifyingGroupBox: TGroupBox;
    trainingOutputButton: TButton;
    testButton: TButton;
    numbOfInputsEdit: TEdit;
    numbOfHiddensEdit: TEdit;
    numbOfHiddens2Edit: TEdit;
    numbOfOutputsEdit: TEdit;
    HiddenLayer2ActivatedCheckBox: TCheckBox;
    numberOfInputsLabel: TLabel;
    NumberOfHiddensLabel: TLabel;
    numberOfHiddens2Label: TLabel;
    numberOfOutputsLabel: TLabel;
    TopologyLabel: TLabel;
    TransFuncHiddLayerComboBox: TComboBox;
    TransFuncHidd2LayerComboBox: TComboBox;
    TransFuncOutputLayerComboBox: TComboBox;
    TransferFunctionsLabel: TLabel;
    TransFuncHiddLayerLabel: TLabel;
    TransFuncHiddLayer2Label: TLabel;
    transFuncOutputLayerLabel: TLabel;
    MomentumEdit: TEdit;
    MomentumLabel: TLabel;
    MLPNet: TTedMLPNet;
    DivergeCheckBox: TCheckBox;
    procedure FormCreate(Sender: TObject);
    procedure learning();
    procedure trainingButtonClick(Sender: TObject);
    procedure LearningButtonClick(Sender: TObject);
    procedure ErrorBoundEditExit(Sender: TObject);
    procedure numbOfCyclesEditExit(Sender: TObject);
    procedure trainingOutputButtonClick(Sender: TObject);
    procedure testButtonClick(Sender: TObject);
    procedure numbOfInputsEditExit(Sender: TObject);
    procedure numbOfHiddensEditExit(Sender: TObject);
    procedure numbOfOutputsEditExit(Sender: TObject);
    procedure numbOfHiddens2EditExit(Sender: TObject);
    procedure HiddenLayer2ActivatedCheckBoxClick(Sender: TObject);
    procedure TransFuncOutputLayerComboBoxChange(Sender: TObject);
    procedure TransFuncHiddLayerComboBoxChange(Sender: TObject);
    procedure TransFuncHidd2LayerComboBoxChange(Sender: TObject);
    procedure MomentumEditExit(Sender: TObject);
  public
    //import training pattern out of file and initialise the net
    procedure trainingAndInitNet();
  private
    FNumberOfLearnCycles: integer;
    FErrorBound: double;

    //gives outputs of the net with training values
    procedure trainingOutput();

    //gives outputs of the net with testdata
    procedure test();
  end;

var
  mlpDemoForm: TmlpDemoForm;

implementation

{$R *.dfm}

procedure TmlpDemoForm.FormCreate(Sender: TObject);
begin
  //initialisation
  DecimalSeparator:='.';
  mlpNet:=TTedMLPnet.create(nil);

  mlpNet.WeightMax:=0.02;
  mlpNet.WeightMin:=-0.02;
  mlpNet.SeedWeights:=310171;
  mlpNet.LearningRateHiddenLayer:=0.05;
  mlpNet.LearningRateOutputLayer:=0.05;
  mlpNet.LearningRateHiddenLayer2:=0.05;

  numbOfInputsEdit.Text:=IntToStr(mlpNet.topology.numberOfInputNeurons);
  numbOfHiddensEdit.Text:=IntToStr(mlpNet.topology.numberOfHiddenNeurons);
  numbOfHiddens2Edit.Text:=IntToStr(mlpNet.topology.numberOfHidden2Neurons);
  numbOfOutputsEdit.Text:=IntToStr(mlpNet.topology.numberOfOutputNeurons);
  HiddenLayer2ActivatedCheckBox.checked:=MLPnet.Topology.HiddenLayer2Active;
  FNumberOfLearnCycles:=1000;
  FErrorBound:=0.1;
  numbOfCyclesEdit.Text:=IntToStr(FNumberOfLearnCycles);
  errorBoundEdit.Text:=FloatToStr(FErrorBound);
  TransFuncHidd2LayerComboBox.ItemIndex:=ord(MLPNet.transFuncHiddLayer2);
  TransFuncHiddLayerComboBox.ItemIndex:=ord(MLPNet.transFuncHiddLayer);
  TransFuncOutputLayerComboBox.ItemIndex:=ord(MLPNet.transFuncOutputLayer);
  MomentumEdit.Text:=FloatToStr(mlpNet.momentum);
end;

procedure TmlpDemoForm.trainingButtonClick(Sender: TObject);
begin
  if openDialog.Execute then
  begin
    trainingAndInitNet();
  end;
end;

procedure TmlpDemoForm.trainingAndInitNet();
var
  numberOfTrainingValues: integer;
  doubles: TDoubleArray;
  messagStr: String;
  trainingInit: boolean;
begin
  ParseTextFileIntoDoubles(openDialog.FileName,doubles,numberOfTrainingValues);
  mlpNet.SetTrainingPatternWithDoubles(doubles,numberOfTrainingValues,
              trainingInit);
  mlpNet.init();
  if mlpNet.isInit() and trainingInit then
  begin
    messagStr:='Trainingspattern parsed and MLP net initialised, ';
    messagStr:=messagStr+'number of training values: '
            +FloatToStr(numberOfTrainingValues);
    ShowMessage(messagStr);
  end;
end;

procedure TmlpDemoForm.LearningButtonClick(Sender: TObject);
begin
   if not mlpNet.isInit() then
   begin
     showMessage('TMLPDemoForm.learningButtonClick: net not initialised');
     exit;
   end;
   learning();
end;

procedure TmlpDemoForm.learning();
var
  dest: TextFile;
  error: double;
  i, cycles: integer;
  cycleStr: String;
  diverging: boolean;
begin
  AssignFile (dest, 'errorStats.txt');
  Rewrite (dest);
  error := 0.0;
  cycles:=1;
  diverging:=false;
  //no errorbound set?
  if not errorCheckBox.checked then
  begin
    i:=0;
    while (i < FNumberOfLearnCycles) and not diverging do
    begin
      mlpNet.learning(cycles,divergeCheckBox.Checked,diverging);
      i:=i+1;
      error := mlpNet.errorOfTrainingData();
      writeln(dest, FloatToStr(error));
    end;
    cycles:=i;
  end
  else // error bounded
  begin
    error := mlpNet.ErrorOfTrainingData();
    i:=0;
    while (i < FNumberOfLearnCycles) and (error > FErrorBound)
    and not diverging do
    begin
      mlpNet.learning(cycles,divergeCheckBox.Checked,diverging,FErrorBound);
      i:=i+1;
      error := mlpNet.errorOfTrainingData();
      writeln(dest, FloatToStr(error));
    end;
    cycles:=i;
  end;
  closeFile(dest);
  if cycles = 1 then
    cycleStr:='cycle'
  else
    cycleStr:='cycles';
  if diverging then showMessage('Net learning has diverged');
  ShowMessage('Learning finished after '+FloatToStr(cycles)+' '+
                        cycleStr+' with error '
                +FloatToStr(error));
end;

procedure TmlpDemoForm.ErrorBoundEditExit(Sender: TObject);
var
  doub: double;
begin
  try
    doub:=StrToFloat(errorBoundEdit.Text);
  except on EConvertError do
  begin
    showMessage('TMLPDemoForm.errorBoundEditOnExit: '
                  +'invalid error bound');
    errorBoundEdit.Text:=FloatToStr(FErrorBound);
    exit;
  end;
  end;
  FErrorBound:=doub;
end;

procedure TmlpDemoForm.numbOfCyclesEditExit(Sender: TObject);
var
  i: integer;
begin
  try
    i:=StrToInt(numbOfCyclesEdit.Text);
  except on EConvertError do
  begin
    showMessage('TMLPDemoForm.numbOfCyclesOnExit: '+
                  'invalid number of learning cycles');
    numbOfCyclesEdit.Text:=IntToStr(FNumberOfLearnCycles);
    exit;
  end;
  end;
  FNumberOfLearnCycles:=i;
end;

procedure TmlpDemoForm.trainingOutputButtonClick(Sender: TObject);
begin
  if not mlpNet.isInit then
  begin
    ShowMessage('TMLPDemoForm.trainingOutputButtonClick:'+
      ' net is not initialised');
    exit;
  end;
  trainingOutput();
end;

procedure TMLPDemoForm.trainingOutput();
var
  error: double;
  i,j: integer;
  outputNet: TDoubleArray;
  trainingPattern: TTrainingPattern;
  inputString, outputString, expectedOutputString: String;
begin
  trainingPattern:=mlpNet.trainingPattern;
  setLength(outputNet,mlpNet.Topology.NumberOfOutputNeurons);
  for i := 0 to trainingPattern.getNumberOfTrainingValues-1 do
  begin
    mlpNet.getOutput(trainingPattern[i].input,outputNet);
    inputString:='';
    for j:=0 to mlpNet.topology.numberOfInputNeurons-1 do
    begin
      inputString:=inputString+'   '+
                        format('%.5f',[trainingPattern[i].input[j]]);
    end;

    expectedOutputString:='';
    outputString:='';
    for j:=0 to mlpNet.topology.numberOfOutputNeurons-1 do
    begin
      expectedOutputString:=expectedOutputString+'   '+
        format('%.5f',[trainingPattern[i].output[j]]);
      error:=abs(trainingPattern[i].output[j]-outputNet[j]);
      outputString:=outputString+'   '+format('%.5f',[outputNet[j]])+
        ' '+'('+format('%.5f',[error])+')';
    end;

    ShowMessage('Training vector no. '+FloatToStr(i+1)+': '+ #13#10 +
                'Input: '+inputString+ #13#10
                +'Expected Output: '+expectedOutputString+#13#10
                +'Real Output (error): '+outputString);
  end;
end;

procedure TmlpDemoForm.test();
var
  doubles, testInput, outputNet: TDoubleArray;
  numberOfTestValues,counter,i,j,numberOfLines: integer;
  inputString, outputString, expectedOutputString: String;
  error: double;
begin
  ParseTextFileIntoDoubles(openDialog.FileName, doubles,numberOfLines);
  //check for invalid datafile
  if (length(doubles) mod (mlpnet.Topology.NumberOfInputNeurons+
                  mlpnet.Topology.NumberOfOutputNeurons)> 0) or
     (length(doubles) div (mlpNet.Topology.NumberOfInputNeurons+
                  mlpNet.Topology.NumberOfOutputNeurons)<>numberOfLines) then
  begin
    ShowMessage('TMLPDemoForm.test: invalid test data');
    exit;
  end;
  numberOfTestValues:=length(doubles) div
        (mlpnet.Topology.NumberOfInputNeurons+
                mlpnet.Topology.NumberOfOutputNeurons);
  counter:=0;
  setLength(testInput,mlpnet.Topology.NumberOfInputNeurons);
  setLength(outputNet,mlpnet.Topology.NumberOfOutputNeurons);
  for i:=0 to numberOfTestValues-1 do
  begin
    inputString:='';
    for j:=0 to mlpnet.Topology.NumberOfInputNeurons-1 do
    begin
      //get Input for the net for ith testvalue
      testInput[j]:=doubles[counter];
      inputString:=inputString+'   '+
                        format('%.5f',[testInput[j]]);
      counter:=counter+1;
    end;
    mlpnet.getOutput(testInput,outputNet);
    expectedOutputString:='';
    outputString:='';
    for j:=0 to mlpnet.Topology.NumberOfOutputNeurons-1 do
    begin
      expectedOutputString:=expectedOutputString+'   '+
        format('%.5f',[doubles[counter]]);
      error:=abs(doubles[counter]-outputNet[j]);
      outputString:=outputString+'   '+format('%.5f',[outputNet[j]])+
        ' '+'('+format('%.5f',[error])+')';
      counter:=counter+1;
    end;
    ShowMessage('Test vector no. '+FloatToStr(i+1)+': '+ #13#10 +
                'Input: '+inputString+ #13#10
                +'Expected Output: '+expectedOutputString+#13#10
                +'Real Output (error): '+outputString);
  end;
end;

procedure TmlpDemoForm.testButtonClick(Sender: TObject);
begin
 if not mlpNet.isInit() then
 begin
   showMessage('TMLPDemoForm.testButtonClick: net is not initialised.');
   exit;
 end;
 if openDialog.Execute then
 begin
   test();
 end;
end;

procedure TmlpDemoForm.numbOfInputsEditExit(Sender: TObject);
var
  int: integer;
begin
  try
    int:=StrToInt(numbOfInputsEdit.Text);
  except on EConvertError do
  begin
    showMessage('TMLPDemoForm.numbOfInputsEditOnExit: invalid topology');
    numbOfInputsEdit.Text:=IntToStr(MLPNet.topology.numberOfInputNeurons);
    exit;
  end;
  end;
  MLPNet.topology.numberOfInputNeurons:=int;
end;

procedure TmlpDemoForm.numbOfHiddensEditExit(Sender: TObject);
var
  int: integer;
begin
  try
    int:=StrToInt(numbOfHiddensEdit.Text);
  except on EConvertError do
  begin
    showMessage('TMLPDemoForm.numbOfHiddensEditOnExit: invalid topology');
    numbOfHiddensEdit.Text:=IntToStr(MLPnet.topology.numberOfHiddenNeurons);
    exit;
  end;
  end;
  MLPnet.topology.numberOfHiddenNeurons:=int;
end;

procedure TmlpDemoForm.numbOfOutputsEditExit(Sender: TObject);
var
  int: integer;
begin
  try
    int:=StrToInt(numbOfOutputsEdit.Text);
  except on EConvertError do
  begin
    showMessage('TMLPDemoForm.numberOfOutputsEditOnExit: invalid topology');
    numbOfOutputsEdit.Text:=IntToStr(MLPnet.topology.numberOfOutputNeurons);
    exit;
  end;
  end;
  MLPnet.topology.numberOfOutputNeurons:=int;
end;

procedure TmlpDemoForm.numbOfHiddens2EditExit(Sender: TObject);
var
  int: integer;
begin
  if MLPnet.Topology.HiddenLayer2Active then
  begin
    try
      int:=StrToInt(numbOfHiddens2Edit.Text);
    except on EConvertError do
    begin
      showMessage('TMLPDemoForm.numbOfHiddens2EditOnExit: invalid topology');
      numbOfHiddens2Edit.Text:=IntToStr(MLPnet.topology.numberOfHidden2Neurons);
      exit;
    end;
    end;
    MLPnet.topology.numberOfHidden2Neurons:=int;
  end
  else
    numbOfHiddens2Edit.Text:='0';
end;

procedure TmlpDemoForm.HiddenLayer2ActivatedCheckBoxClick(Sender: TObject);
begin
  MLPNet.Topology.HiddenLayer2Active:=not MLPNet.Topology.HiddenLayer2Active;
  numbOfHiddens2Edit.Text:='0';
  MLPNet.Topology.numberOfHidden2Neurons:=0;
end;

procedure TmlpDemoForm.TransFuncOutputLayerComboBoxChange(Sender: TObject);
begin
  if (TransFuncOutputLayerComboBox.ItemIndex in [0..2]) then
    mlpnet.transFuncOutputLayer:=
          TTransferFunction(TransFuncOutputLayerComboBox.ItemIndex);

end;

procedure TmlpDemoForm.TransFuncHiddLayerComboBoxChange(Sender: TObject);
begin
  if (TransFuncHiddLayerComboBox.ItemIndex in [0..2]) then
    mlpnet.transFuncHiddLayer:=
          TTransferFunction(TransFuncHiddLayerComboBox.ItemIndex);

end;

procedure TmlpDemoForm.TransFuncHidd2LayerComboBoxChange(Sender: TObject);
begin
  if (TransFuncHidd2LayerComboBox.ItemIndex in [0..2]) and
      mlpnet.Topology.HiddenLayer2active then
    mlpnet.transFuncHiddLayer2:=
          TTransferFunction(TransFuncHidd2LayerComboBox.ItemIndex)
  else
    TransFuncHidd2LayerComboBox.ItemIndex:=2;

end;

procedure TmlpDemoForm.MomentumEditExit(Sender: TObject);
var
  doub: double;
begin
  try
    doub:=StrToFloat(MomentumEdit.Text);
  except on EConvertError do
  begin
    showMessage('TMLPDemoForm.MomentumEditOnExit: '
                  +'invalid momentum');
    MomentumEdit.text:=FloatToStr(mlpNet.Momentum);
    exit;
  end;
  end;
  if (doub < 0) or (doub >=1) then
  begin
    showMessage('TMLPDemoForm.MomentumEditOnExit: '
                  +'invalid momentum');
    MomentumEdit.text:=FloatToStr(mlpNet.Momentum);
    exit;
  end
  else
    mlpNet.Momentum:=doub;
end;


end.
